#!/usr/bin/python
#
# Copyright (c) 2012, Stuart Prescott
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
###

"""
Gather the list of debian.net domains that are in the debian LDAP and
print out either a plain text table or an HTML table of the domain names.

Prerequisite packages:

    python-ldap

"""

import sys
import ldap
import encodings.idna
import logging
from operator import attrgetter

class DebianDotNetDomain(object):
    """A single domain name entry from LDAP"""
    known_types = ['A', 'AAAA', 'CNAME', 'MX', 'TXT']
    def __init__(self, line, owner):
        """ Create a domain ownership record with an associated owner

        line: the raw text of the dnsZoneEntry line from LDAP as described
              in http://wiki.debian.org/DebianDotNet
        owner: a DebianLdapEntry object that is the owner of this dnsZoneEntry
        """
        l = line.split()
        self.name = l[0]
        self.type = l[2].upper()
        if self.type not in self.known_types:
            logging.warn("Unknown dns entry type '%s' in record: %s",
                          self.type, line)
        if self.type in [ 'CNAME', 'A', 'AAAA']:
            self.target = l[3]
        elif self.type == 'TXT':
            self.target = " ".join(l[3:])  ## TODO: check this format
        elif self.type == 'MX':
            self.target = l[4]
        if self.name.startswith("xn--"):
            self.realname = encodings.idna.ToUnicode(self.name)
        else:
            self.realname = self.name
        self.owner = owner
        self.description = ""


class DebianLdapEntry(object):
    """ An entry in the Debian directory """
    def __init__(self, dn, entry):
        """ create an object from an ldap search result """
        self.dn = dn
        self.uid = entry['uid'][0]
        self.gecos = entry['gecos'][0].decode('utf-8')
        self.name = self.gecos.split(',')[0]

        self.dnsZoneEntry = entry['dnsZoneEntry']
        self.domains = []
        for entry in self.dnsZoneEntry:
            self.domains.append(DebianDotNetDomain(entry, self))


class DebianDotNetDomains(dict):
    """ Container a collection of DebianDotNetDomain entries """
    known_formats = ['text', 'html', 'moin', 'wmldata']
    def __init__(self, domain_entries):
        """ load a list of DebianDotNetDomain objects """
        super(dict, self).__init__()
        self.format = 'text'
        self.descriptions = {}
        self.add(domain_entries)

    def add(self, domain_entries):
        """ incorporate a list of DebianDotNetDomain objects """
        for domain in domain_entries:
            if domain.realname not in self:
                self[domain.realname] = domain
            else:
                # item is already in the dict; merge the records appropriately
                if domain.type == 'TXT':
                    self.descriptions[domain.realname] = domain.target
                elif domain.type == self[domain.realname].type and \
                      domain.type in ['CNAME', 'A', 'AAAA']:
                    ## FIXME handle A vs AAAA properly
                    logging.warn("Ignoring extra records for domain '%s'. ",
                                  domain.realname)
        self._descriptions_merge()

    def merge_descriptions_file(self, description_filename):
        """ merge in the old static descriptions stored in a text file

        description_filename: text file with descriptions. File format is:
            domain    descrription
        where any amount of whitespace around the "domain" part is permitted.
        The ".debian.net" part of the domain should be omitted. Any domain
        descriptions already provided by a TXT record are ignored.
        """
        for line in file(description_filename):
            (domain, description) = line.split(None, 1)
            if domain not in self.descriptions:
                self.descriptions[domain.strip()] = description.strip()
            else:
                logging.info("Ignoring duplicate description for domain %s.",
                              domain)
        self._descriptions_merge()

    def _descriptions_merge(self):
        """ copy the descriptions into the DebianDotNetDomain entries """
        for (domain, description) in self.descriptions.items():
            if domain in self and not self[domain].description:
                self[domain].description = description

    def __str__(self):
        """ make a string representation according to the desired format """
        return {
              'text': self.as_text,
              'html': self.as_html,
              'moin': self.as_moin,
              'wmldata': self.as_wmldata
            }[self.format]()

    def as_html(self):
        """ format the entries in an HTML table """
        tableformat = "<tr><td>%(realname)s</td><td>%(description)s</td>"\
                      "<td><a href='https://db.debian.org/search.cgi?"\
                      "dosearch=1;uid=%(uid)s'>"\
                      "%(ownername)s (%(uid)s)</a></td></tr>"
        s = [ "<table><tr><th>Domain</td>"\
              "<th>Description</th><th>Owner</th></tr>" ]
        s.append(self.as_format(tableformat))
        s.append("</table>")
        return "\n".join(s)

    def as_wmldata(self):
        """ format the entries in a wml data file """
        tableformat = "<item>\n"\
                      "  <name %(realname)s>\n"\
                      "  <description %(description)s>\n"\
                      "  <ownername %(ownername)s>\n"\
                      "  <uid %(uid)s>\n"\
                      "</item>\n"
        s = [ '#include "$(ENGLISHDIR)/misc/debiandotnet.def"',
              '<domainlist>' ]
        s.append(self.as_format(tableformat))
        s.append('</domainlist>')
        return "\n".join(s)

    def as_text(self):
        """ format the entries in fixed-width columns """
        tableformat = "%(realname)-25s%(description)-50s%(ownername)-40s%(uid)s"
        s = [ tableformat % \
                        { 'realname': 'Domain',
                          'description': 'Description',
                          'uid' : 'UID',
                          'ownername' : 'Name' } ]
        s.append(self.as_format(tableformat))
        return "\n".join(s)

    def as_moin(self):
        """ format the entries in a moinmoin wiki format table """
        tableformat = "||%(realname)s|| %(description)s || "\
                  "[[https://db.debian.org/search.cgi?dosearch=1;uid=%(uid)s'"\
                  "|%(ownername)s (%(uid)s)]] ||"
        s = [ "||'''alias'''|| '''description''' "\
                "||'''comment'''|| '''admin'''||" ]
        s.append(self.as_format(tableformat))
        return "\n".join(s)

    def as_format(self, tableformat):
        """ format the entries according to the given format string """
        ## FIXME: html encoding of output?
        s = []
        for domain in sorted(self.values(), key=attrgetter('realname')):
            if domain.type in ['CNAME', 'A', 'AAAA']:
                s.append( (tableformat %\
                            { 'realname': domain.realname,
                              'description': domain.description,
                              'uid' : domain.owner.uid,
                              'ownername' : domain.owner.name })\
                            .encode('UTF-8') )
        return "\n".join(s)


def find_ddn(server, base, domain):
    """ find all debian.net domains matching the designated pattern

      server: ldap server to contact (machine name or ip)
      base: the BaseDN to use for the query
      domain: the domain to search for (or * to search for all domains)

      returns: list of DebianDotNetDomain objects
    """
    scope = ldap.SCOPE_SUBTREE
    # See
    #   http://anonscm.debian.org/gitweb/?p=mirror/userdir-ldap.git;a=blob;f=UDLdap.py;hb=HEAD
    #   http://anonscm.debian.org/gitweb/?p=mirror/userdir-ldap.git;a=blob;f=ud-generate;hb=HEAD
    # for details of how valid records are found.
    lfilter = "(&(dnsZoneEntry=%s)(|(!(accountStatus=*))(accountStatus=active)(objectClass=debianRoleAccount)))" % domain
    retrieve_attributes = None
    domains = []

    try:
        logging.info("Binding to server %s.", server)
        connection = ldap.open(server)
        logging.info("Search for '%s' in baseDN '%s'", lfilter, base)
        for (dn, attrs) in connection.search_s(base, scope,
                                      lfilter, retrieve_attributes):
            e = DebianLdapEntry(dn, attrs)
            domains.extend(e.domains)
    except ldap.LDAPError, e:
        logging.error("Error extracting data from directory: %s", str(e))
    return domains

def find_ddn_dummy(server, base, domain):
    """ Load some dummy entries into the collection for testing """
    records = [
              {"uid": ["foo"], "gecos": ["Foo Bar,,,,"],
                  "dnsZoneEntry": ["foo IN A 1.2.3.4"]},
              {"uid": ['goo'], "gecos": ["Goo Bar,,,,"],
                  "dnsZoneEntry": ["goo IN CNAME goo.example.net."]},
              {"uid": ["hoo"], "gecos": ["Hoo Bar,,,,"],
                  "dnsZoneEntry": ["hoobar IN CNAME hoobar.example.net.",
                              "hoohoo IN CNAME hoohoo.example.org.",
                              "hoohar IN A 6.7.8.9",
                              "hoohar IN MX 10 hoohar.example.net.",
                              "hoohar IN TXT 'test data for ddn'"]}
          ]

    domains = []
    for attrs in records:
        e = DebianLdapEntry("", attrs)
        domains.extend(e.domains)
    return domains


def get_domain_list(server, base, domain, description_filename):
    """ obtain a list of all debian.net domains """
    domains = DebianDotNetDomains(find_ddn(server, base, domain))
    #domains = DebianDotNetDomains(find_ddn_dummy(server, base, domain))
    if not domains:
        logging.error("No results were obtained from the directory.")
        return
    domains.merge_descriptions_file(description_filename)
    return domains

def main():
    """ obtain a list of all debian.net domains and print it to STDOUT """
    logging.basicConfig()
    server = "db.debian.org"
    base = "dc=debian,dc=org"
    domain = '*'
    description_filename = 'debiandotnet.descriptions'

    domains = get_domain_list(server, base, domain, description_filename)
    if not domains:
        sys.exit(-1)

    output_format = DebianDotNetDomains.known_formats[0]
    if len(sys.argv) > 1 and sys.argv[1] in DebianDotNetDomains.known_formats:
        output_format = sys.argv[1]
    domains.format = output_format
    print str(domains)


if __name__ == '__main__':
    main()
